package agent

import (
	"context"
	"fmt"
	"log"
	"strconv"
	"sync"
	"time"

	"github.com/it-novum/nsta/gearman"
	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

// CommandBackend should implement the execution of commands on the remote machine
type CommandBackend interface {
	CmkRunCommand(ctx context.Context, hostUUID string, args []string, env map[string]string) (string, error)
	CmkWriteFile(ctx context.Context, hostUUID, path string, content []byte) error
	QuoteCommandArg(string) string
}

// Agent manages the execution of openITCOCKPIT Monitoring Agent commands
type Agent struct {
	mtx           sync.Mutex
	SatelliteName string
}

// Send and HTTP/S request to the openITCOCKPIT Monitoring Agent to get the latest json output
func (a *Agent) queryAgent(cmd CommandBackend, msg *gearman.OitcAgentMessage, response *gearman.OitcAgentResponse) {
	var err error

	ctx, cancel := context.WithTimeout(context.Background(), time.Second*120)
	defer cancel()

	env := map[string]string{}

	log.Println(&Error{Name: a.SatelliteName, HostUUID: msg.Data.Task, Message: "Query JSON from openITCOCKPIT Monitoring Agent"})

	var queryComand []string

	// Use Auto-TLS by default
	queryComand = []string{"/opt/openitc/receiver/bin/poller.php", "query", "-H", msg.Data.Address, "-p", strconv.Itoa(msg.Data.Port), "-C", "/opt/openitc/agent/server_ca.pem", "-K", "/opt/openitc/agent/server_ca.key", "--auto-tls"}

	if !msg.Data.UseAutossl && msg.Data.UseHttps {
		// Use HTTPS
		queryComand = []string{"/opt/openitc/receiver/bin/poller.php", "query", "-H", msg.Data.Address, "-p", strconv.Itoa(msg.Data.Port), "-C", "/opt/openitc/agent/server_ca.pem", "-K", "/opt/openitc/agent/server_ca.key", "--https"}
		if !msg.Data.UseAutossl && !msg.Data.Insecure {
			queryComand = append(queryComand, "-k")
		}
	}

	if !msg.Data.UseAutossl && !msg.Data.UseHttps {
		// HTTP only
		queryComand = []string{"/opt/openitc/receiver/bin/poller.php", "query", "-H", msg.Data.Address, "-p", strconv.Itoa(msg.Data.Port), "-C", "/opt/openitc/agent/server_ca.pem", "-K", "/opt/openitc/agent/server_ca.key", "--http"}
	}

	if msg.Data.BasicAuth {
		queryComand = append(queryComand, "--username")
		queryComand = append(queryComand, msg.Data.Username)
		queryComand = append(queryComand, "--password")
		queryComand = append(queryComand, msg.Data.Password)
	}

	if msg.Data.TestConnectionOnly {
		queryComand = append(queryComand, "--test-connection-and-autotls")
	}

	if msg.Data.AutosslSuccessful {
		queryComand = append(queryComand, "--autossl-successful")
	}

	if response.RawResult, err = cmd.CmkRunCommand(ctx, msg.Data.Task, queryComand, env); err != nil {
		response.Error = fmt.Sprint("could not run agent receiver: ", err)
		log.Println(err)
		return
	}

	log.Println(&Error{Name: a.SatelliteName, HostUUID: msg.Data.Task, Message: "successful agent query"})
}

// Get a list of avilable oITC Agents running in Push mode for "select_agent" step in Wizard
func (a *Agent) getPushAgents(cmd CommandBackend, msg *gearman.OitcAgentMessage, response *gearman.OitcAgentResponse) {
	var err error

	ctx, cancel := context.WithTimeout(context.Background(), time.Second*120)
	defer cancel()

	env := map[string]string{}

	log.Println(&Error{Name: a.SatelliteName, HostUUID: msg.Data.Task, Message: "Query JSON from openITCOCKPIT Monitoring Agent"})

	// Execute the oITC Agent CLI API endpoint
	queryComand := []string{"oitc", "agent", "--get-push-agents"}
	if msg.Data.HostUuid != "" {
		queryComand = append(queryComand, "--host-uuid")
		queryComand = append(queryComand, msg.Data.HostUuid)
	}

	if response.RawResult, err = cmd.CmkRunCommand(ctx, msg.Data.Task, queryComand, env); err != nil {
		response.Error = fmt.Sprint("could not run oITC agent CLI API command: ", err)
		log.Println(err)
		return
	}

	log.Println(&Error{Name: a.SatelliteName, HostUUID: msg.Data.Task, Message: "successful executed oITC Agent CLI API endpoint"})
}

// Get a list of all Push Agents (used by Import Module)
func (a *Agent) getAllPushAgents(cmd CommandBackend, msg *gearman.OitcAgentMessage, response *gearman.OitcAgentResponse) {
	var err error

	ctx, cancel := context.WithTimeout(context.Background(), time.Second*120)
	defer cancel()

	env := map[string]string{}

	log.Println(&Error{Name: a.SatelliteName, HostUUID: msg.Data.Task, Message: "Query all oITC Agents running in push mode"})

	// Execute the oITC Agent CLI API endpoint
	queryComand := []string{"oitc", "agent", "--get-all-push-agents"}

	if response.RawResult, err = cmd.CmkRunCommand(ctx, msg.Data.Task, queryComand, env); err != nil {
		response.Error = fmt.Sprint("could not run oITC agent CLI API command: ", err)
		log.Println(err)
		return
	}

	log.Println(&Error{Name: a.SatelliteName, HostUUID: msg.Data.Task, Message: "successful executed oITC Agent CLI API endpoint"})
}

// Assign a HostUUID to an Agent
func (a *Agent) assignHost(cmd CommandBackend, msg *gearman.OitcAgentMessage, response *gearman.OitcAgentResponse) {
	var err error

	ctx, cancel := context.WithTimeout(context.Background(), time.Second*120)
	defer cancel()

	env := map[string]string{}

	log.Println(&Error{Name: a.SatelliteName, HostUUID: msg.Data.Task, Message: "Query JSON from openITCOCKPIT Monitoring Agent"})

	// Execute the oITC Agent CLI API endpoint
	queryComand := []string{"oitc", "agent", "--assign", "--push-agent-id", strconv.Itoa(msg.Data.PushAgentId), "--host-uuid", msg.Data.HostUuid}

	if response.RawResult, err = cmd.CmkRunCommand(ctx, msg.Data.Task, queryComand, env); err != nil {
		response.Error = fmt.Sprint("could not assign agent to host: ", err)
		log.Println(err)
		return
	}

	log.Println(&Error{Name: a.SatelliteName, HostUUID: msg.Data.Task, Message: "successful assigned agent to host."})
}

// Query Agent Output from an Agent running in Push Mode
func (a *Agent) queryPushAgent(cmd CommandBackend, msg *gearman.OitcAgentMessage, response *gearman.OitcAgentResponse) {
	var err error

	ctx, cancel := context.WithTimeout(context.Background(), time.Second*120)
	defer cancel()

	env := map[string]string{}

	log.Println(&Error{Name: a.SatelliteName, HostUUID: msg.Data.Task, Message: "Query JSON from openITCOCKPIT Monitoring Agent"})

	// Execute the oITC Agent CLI API endpoint
	queryComand := []string{"oitc", "agent", "--query-agent", "--host-uuid", msg.Data.HostUuid}

	if response.RawResult, err = cmd.CmkRunCommand(ctx, msg.Data.Task, queryComand, env); err != nil {
		response.Error = fmt.Sprint("could not assign agent to host: ", err)
		log.Println(err)
		return
	}

	log.Println(&Error{Name: a.SatelliteName, HostUUID: msg.Data.Task, Message: "successful assigned agent to host."})
}

// ProcessMessage parses a oITC Agent message and runs the required commands on the satellite
func (a *Agent) ProcessMessage(cmd CommandBackend, msg *gearman.Message) *gearman.OitcAgentResponse {
	a.mtx.Lock()
	defer a.mtx.Unlock()

	m := &gearman.OitcAgentMessage{}
	err := json.Unmarshal([]byte(msg.String), m)

	response := &gearman.OitcAgentResponse{
		SatelliteID:   m.SatelliteID,
		Command:       m.Command,
		TaskID:        m.Data.TaskID,
		AgentConfigId: m.Data.AgentConfigId,
	}
	if err != nil {
		log.Println(&Error{Name: a.SatelliteName, Message: "could not parse message", Err: err})
		response.Error = fmt.Sprint("could not parse message for oITC Agent:", err)
		return response
	}

	switch m.Data.Task {
	case "query":
		a.queryAgent(cmd, m, response)
	case "get_push_agents":
		a.getPushAgents(cmd, m, response)
	case "get_all_push_agents":
		a.getAllPushAgents(cmd, m, response)
		response.IsGetAllPushAgentsRequest = true
	case "assign_host":
		a.assignHost(cmd, m, response)
	case "query_push":
		a.queryPushAgent(cmd, m, response)
	default:
		log.Println(&Error{Name: a.SatelliteName, Message: fmt.Sprint("unknown task: ", m.Data.Task)})
		response.Error = fmt.Sprint("unknown task for oITC Agent: ", m.Data.Task)
	}

	return response
}
